/*
 * Copyright © 2018 Lisk Foundation
 *
 * See the LICENSE file at the top-level directory of this distribution
 * for licensing information.
 *
 * Unless otherwise agreed in a custom licensing agreement with the Lisk Foundation,
 * no part of this software, including this file, may be copied, modified,
 * propagated, or distributed except according to the terms contained in the
 * LICENSE file.
 *
 * Removal or modification of this copyright notice is prohibited.
 */

'use strict';

const _ = require('lodash');
const swaggerHelper = require('../../helpers/swagger');
const ApiError = require('../../helpers/api_error');

// Private Fields
let modules;

/**
 * Description of the function.
 *
 * @class
 * @memberof api.controllers
 * @requires lodash
 * @requires helpers/apiError
 * @requires helpers/swagger.generateParamsErrorObject
 * @requires helpers/swagger.invalidParams
 * @param {Object} scope - App instance
 * @todo Add description of TransactionsController
 */
function TransactionsController(scope) {
	modules = scope.modules;
}

/**
 * Description of the function.
 *
 * @param {Object} context
 * @param {function} next
 * @todo Add description for the function and the params
 */
TransactionsController.getTransactions = function(context, next) {
	const invalidParams = swaggerHelper.invalidParams(context.request);

	if (invalidParams.length) {
		return next(swaggerHelper.generateParamsErrorObject(invalidParams));
	}

	const params = context.request.swagger.params;

	let filters = {
		senderIdOrRecipientId: params.senderIdOrRecipientId.value,
		id: params.id.value,
		blockId: params.blockId.value,
		recipientId: params.recipientId.value,
		recipientPublicKey: params.recipientPublicKey.value,
		senderId: params.senderId.value,
		senderPublicKey: params.senderPublicKey.value,
		type: params.type.value,
		fromHeight: params.height.value,
		toHeight: params.height.value,
		fromTimestamp: params.fromTimestamp.value,
		toTimestamp: params.toTimestamp.value,
		minAmount: params.minAmount.value,
		maxAmount: params.maxAmount.value,
		data: params.data.value,
		sort: params.sort.value,
		limit: params.limit.value,
		offset: params.offset.value,
	};

	// Remove filters with null values
	filters = _.pickBy(filters, v => !(v === undefined || v === null));

	return modules.transactions.shared.getTransactions(
		_.clone(filters),
		(err, data) => {
			if (err) {
				return next(err);
			}

			const transactions = _.map(
				_.cloneDeep(data.transactions),
				transaction => {
					transaction.senderId = transaction.senderId || '';
					transaction.recipientId = transaction.recipientId || '';
					transaction.recipientPublicKey = transaction.recipientPublicKey || '';

					transaction.amount = transaction.amount.toString();
					transaction.fee = transaction.fee.toString();

					return transaction;
				}
			);

			return next(null, {
				data: transactions,
				meta: {
					offset: filters.offset,
					limit: filters.limit,
					count: parseInt(data.count),
				},
			});
		}
	);
};

/**
 * Description of the function.
 *
 * @param {Object} context
 * @param {function} next
 * @todo Add description for the function and the params
 */
TransactionsController.postTransaction = function(context, next) {
	const transaction = context.request.swagger.params.transaction.value;

	return modules.transactions.shared.postTransaction(
		transaction,
		(err, data) => {
			if (err) {
				if (err instanceof ApiError) {
					context.statusCode = err.code;
					delete err.code;
				}

				return next(err);
			}

			return next(null, {
				data: { message: data },
				meta: { status: true },
			});
		}
	);
};

module.exports = TransactionsController;
