'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});

var _url = require('url');

var _url2 = _interopRequireDefault(_url);

var _liskCryptography = require('@liskhq/lisk-cryptography');

var _liskCryptography2 = _interopRequireDefault(_liskCryptography);

var _base = require('../../base');

var _base2 = _interopRequireDefault(_base);

var _constants = require('../../utils/constants');

var _error = require('../../utils/error');

var _config = require('../../utils/config');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*
 * LiskHQ/lisk-commander
 * Copyright © 2017–2018 Lisk Foundation
 *
 * See the LICENSE file at the top-level directory of this distribution
 * for licensing information.
 *
 * Unless otherwise agreed in a custom licensing agreement with the Lisk Foundation,
 * no part of this software, including this file, may be copied, modified,
 * propagated, or distributed except according to the terms contained in the
 * LICENSE file.
 *
 * Removal or modification of this copyright notice is prohibited.
 *
 */
const availableVariables = _constants.CONFIG_VARIABLES.join(', ');

const WRITE_FAIL_WARNING = 'Config file could not be written: your changes will not be persisted.';

const NETHASH_ERROR_MESSAGE = 'Value must be a hex string with 64 characters, or one of main or test.';

const URL_ERROR_MESSAGE = `Node URLs must include a supported protocol (${_constants.API_PROTOCOLS.map(protocol => protocol.replace(':', '')).join(', ')}) and a hostname. E.g. https://127.0.0.1:4000 or http://localhost.`;

const checkBoolean = value => ['true', 'false'].includes(value);

const setNestedConfigProperty = (config, path, value) => {
	const dotNotationArray = path.split('.');
	dotNotationArray.reduce((obj, pathComponent, i) => {
		if (i === dotNotationArray.length - 1) {
			if (obj === undefined) {
				throw new _error.ValidationError(`Config file could not be written: property '${dotNotationArray.join('.')}' was not found. It looks like your configuration file is corrupted. Please check the file at ${process.env.XDG_CONFIG_HOME} or remove it (a fresh default configuration file will be created when you run Lisk Commander again).`);
			}
			// eslint-disable-next-line no-param-reassign
			obj[pathComponent] = value;
			return config;
		}
		return obj[pathComponent];
	}, config);
};

const attemptWriteToFile = (newConfig, value, dotNotation) => {
	const writeSuccess = (0, _config.setConfig)(process.env.XDG_CONFIG_HOME, newConfig);

	if (!writeSuccess) {
		throw new _error.FileSystemError(WRITE_FAIL_WARNING);
	}

	const message = value === '' || Array.isArray(value) && value.length === 0 ? `Successfully reset ${dotNotation}.` : `Successfully set ${dotNotation} to ${value}.`;

	const result = {
		message
	};

	return result;
};

const setValue = (config, dotNotation, value) => {
	setNestedConfigProperty(config, dotNotation, value);
	return attemptWriteToFile(config, value, dotNotation);
};

const setBoolean = (config, dotNotation, value) => {
	if (!checkBoolean(value)) {
		throw new _error.ValidationError('Value must be a boolean.');
	}
	const newValue = value === 'true';
	return setValue(config, dotNotation, newValue);
};

const setArrayURL = (config, dotNotation, value, inputs) => {
	inputs.forEach(input => {
		const { protocol, hostname } = _url2.default.parse(input);
		if (!_constants.API_PROTOCOLS.includes(protocol) || !hostname) {
			throw new _error.ValidationError(URL_ERROR_MESSAGE);
		}
	});
	return setValue(config, dotNotation, inputs);
};

const setNethash = (config, dotNotation, value) => {
	if (dotNotation === 'api.network' && !Object.keys(_constants.NETHASHES).includes(value)) {
		if (value.length !== 64) {
			throw new _error.ValidationError(NETHASH_ERROR_MESSAGE);
		}
		try {
			_liskCryptography2.default.hexToBuffer(value, 'utf8');
		} catch (error) {
			throw new _error.ValidationError(NETHASH_ERROR_MESSAGE);
		}
	}
	return setValue(config, dotNotation, value);
};

const handlers = {
	'api.nodes': setArrayURL,
	'api.network': setNethash,
	json: setBoolean,
	name: setValue,
	pretty: setBoolean
};

class SetCommand extends _base2.default {
	async run() {
		const { args: { variable, values } } = this.parse(SetCommand);
		const safeValues = values || [];
		const safeValue = safeValues[0] || '';
		const result = handlers[variable](this.userConfig, variable, safeValue, safeValues);
		this.print(result, true);
	}
}

exports.default = SetCommand;
SetCommand.args = [{
	name: 'variable',
	required: true,
	options: _constants.CONFIG_VARIABLES,
	description: ''
}, {
	name: 'values',
	required: false,
	parse: input => input.split(','),
	description: ''
}];

SetCommand.flags = Object.assign({}, _base2.default.flags);

SetCommand.description = `
Sets configuration.
...
Variables available: ${availableVariables}.
`;

SetCommand.examples = ['config:set json true', 'config:set api.network main', 'config:set api.nodes https://127.0.0.1:4000,http://mynode.com:7000'];